/*
 * RASP89 - Rocket Altitude Simulation Program
 * Based on the simple BASIC program written by
 * Harry Stine in 1979, modified and converted
 * into C by Mark A. Storin, Sept/Oct 1989.
 * Rev 2 by Kent Hoult, June 1990.
 *
 * 10-20-89  modified to allow multiple stages up to 3.
 * 07-01-90  modified to use linear segment engine thrust tables from data file
 *           smaller time increments, and crude supersonic transition drag.
 * 6/7/93    Modified to support staging delays.
 *           Made changes to the supersonic transition drag equations.
 *           Migrated lots of global variables to local (with meaninful names)
 *           Rev #3  Stu Barrett
 * 12/29/93  Added support for VAX/VMS - Rusty Whitman
 * 02/15/94  Decrease air density with altitude/temp - added air_density()
 *           Make output gnuplot compatible 
 *           Rev 3.2  - Mark C. Spiegl (spiegl@rtsg.mot.com) 
 */

#define UNIX   1
/* #define MSDOS  1 */
/* #define VMS    1 */ 

#include <stdio.h>
#include <ctype.h>

#define VERSION   "3.2" 
#define TRUE      1
#define FALSE     0
#define MAXENG    3
#define MAXT      64
#define MAXSTAGE  3
#define PI        3.1415927
#define G         9.80621
#define IN2M      0.0254
#define GM2LB     0.00220462
#define OZ2KG     0.028349523
#define M2FT      3.280840
#define ROD       1.2192      /* Length of launch rod in meters (48") */
#define MACH0_8   265.168     /* Mach 0.8 in Meters/sec  870 ft/sec */
#define MACH1     331.460     /* Mach 1.0 in Meters/sec  1086 ft/sec*/
#define MACH1_2   397.752     /* Mach 1.2 in Meters/sec  1305 ft/sec */

/*
 * Define the 1st character on non-data output lines. This forces 
 * output to be gnuplot compatible. 
 */ 
#ifdef VMS
char ch1 = '!'; 
#else
char ch1 = '#'; 
#endif

/*
 * Global Variables
 */
int verbose = TRUE;
double rocketwt;
double drag_coff;
double base_temp; 

int
  nexteng,      /* next free engine index */
  destnum,      /* index into array of devices for output - dest */
  stagenum;     /* number of stages */

char Mcode[16];   /* array to hold engine code */
char fname[64];   /* output file name for simulation */

FILE *stream,  *efile, *fopen();

char efile_name[] = {"rasp.eng"}; /* name of engine database file */

struct engine {
    double t2;         /* thrust duration    (seconds)                 */
    double m2;         /* propellant wt.     (kilograms)               */
    double wt;         /* initial engine wt. (kilograms)               */
    double thrust[MAXT]; /* thrust curve     (Time-sec,Thrust-Newtons) */
    double navg;       /* average thrust     (newtons)                 */
    double ntot;       /* total impulse      (newton-seconds)          */
    double npeak;      /* peak thrust        (newtons)                 */
    int   dia;         /* engine diameter    (millimeters)             */
    int   len;         /* engine length      (millimeters)             */
    int   delay[8];    /* ejection delays available (sec)              */
    char  mfg[32];     /* manufacturer info                            */
    char  code[32];    /* engine name                                  */
};

struct stage_info {
    int engcnum;      /* index into e_info */
    int engnum;       /* # of engines in stage */
    double start_burn; /* Start of engine (includes previous stage) */
    double end_burn;   /* End of engine burn  (includes previous stage) */
    double end_stage;  /* End of stage (includes previous stage) */
    double weight;    /* stage wt w/o engine */
    double totalw;    /* wt of stage w/ engine(s) */
    double maxd;      /* max diameter of stage  */
};
struct stage_info stages[MAXSTAGE];
struct engine e_info[MAXENG];


print_engine_header() {
  fprintf(stream,"%c\n%c\t\tAverage\tTotal\tPeak\tBurn\n", ch1, ch1);
  fprintf(stream,"%c\t\tThrust\tImpulse\tThrust\tTime\tDelays\n", ch1);
  fprintf(stream,"%cType  Mfg/Opt\t (Nt)\t(Nt-Sec) (Nt)\t(Sec)\t(Sec)\n", ch1);
  fprintf(stream,"%c----  -------\t------\t-------\t------\t-----\t------\n", ch1);
}

print_engine_info(e)
int e;
{
  int d;
  print_engine_header();
  fprintf(stream,"%c%-6s %-6s\t%.2lf\t%.2lf\t%.2lf\t%.2lf\t", ch1, 
         e_info[e].code, e_info[e].mfg, e_info[e].navg, e_info[e].ntot,
         e_info[e].npeak, e_info[e].t2);
  for(d = 0; e_info[e].delay[d] >= 0; d++)
    fprintf(stream,"%d ", e_info[e].delay[d]);
  fprintf(stream,"\n");
}

load_engine(e) int e; {
  int t, dia, len, more;
  char s[256], *c, sdelay[32];
  double sum,peak, t1, t2, f1, f2;
  double *p;

  for(more = TRUE; more;) {
    if((c = fgets(s, 256, efile)) != NULL) {
      if(*c != ';') {
        if(sscanf(s, "%s %d %d %s %lf %lf %s", e_info[e].code, &e_info[e].dia,
                  &e_info[e].len, sdelay, &e_info[e].m2, &e_info[e].wt,
                  e_info[e].mfg) != 7) {
          printf("\n*** Error - Bad line in rasp.eng\n");
          exit();
        }
        more = FALSE;
      }
    }
    else
      return(-1);
  }
  t = 0;
  for(more = TRUE; more;) {
    if((c = fgets(s, 256, efile)) != NULL) {
      if(*c != ';') {
        if(t >= MAXT) {
          printf("\n*** Error - Thrust table overflow in rasp.eng\n");
          exit();
        }
        if(sscanf(s, "%lf %lf",
                  &e_info[e].thrust[t], &e_info[e].thrust[t+1]) != 2) {
          printf("\n*** Error - Bad thrust table line in rasp.eng\n");
          exit();
        }
        if(e_info[e].thrust[t+1] == 0) {
          e_info[e].t2 = e_info[e].thrust[t];
          more = FALSE;
        }
        else
          t += 2;
      }
    }
    else {
      printf("\n*** Error - unexpected EOF while loading thrust curve.\n");
      exit();
    }
  }
  t1 = f1 = sum = peak = 0.0;
  p = &e_info[e].thrust[0];
  do {
    t2 = *p++;
    f2 = *p++;
    if(f2 > peak) peak = f2;
    sum += (t2 - t1) * (f1 + f2) * 0.5;
    t1 = t2;
    f1 = f2;
  } while (f1 != 0.0);
  e_info[e].ntot = sum;
  e_info[e].npeak = peak;
  e_info[e].navg = sum / t1;
  c = sdelay;
  for(t = 0; t < 7; t++) {
    e_info[e].delay[t] = (int) strtol(c, &c, 10);
    if(*c != '-') {
      e_info[e].delay[t+1] = -1;
      break;
    }
    else
      c++;

  }
  return(e);
}

int getmotor(st) int st; {
    int i, match, e;

    match = FALSE;
    e = nexteng++;
    do {
        if (stagenum > 1)
        printf("Stage %d ", st);
        printf("Motor Code? ");
        scanf("%s", Mcode);
        for (i = 0; i < strlen(Mcode); i++)
            if(islower(Mcode[i]))
               Mcode[i] = toupper(Mcode[i]);
        if((efile = fopen(efile_name,"r")) == NULL) {
          printf("\n*** Error - unable to open engine database %s ***\n",
                 efile_name);
          exit();
        }
        while(load_engine(e) >= 0) {
          if (strcmp(Mcode, e_info[e].code) == 0) {
            match = TRUE;
            break;
          }
        }
      } while (match == FALSE);
    fclose(efile);
    return(e);
  }


choices() {
    int i;
    double wt;
    double stage_delay = 0;

    printf("Number of stages? ");
    scanf("%d", &stagenum);
    for (i = 0; i < stagenum; i++) {
        stages[i].engcnum = getmotor(i + 1);
        if (stagenum > 1)
            printf("Weight of stage %d w/o engine in ounces? ", i+1);
        else
            printf("Weight of rocket w/o engine in ounces? ");
        scanf("%lE", &stages[i].weight);
        stages[i].weight *= OZ2KG;
        if (stagenum > 1)
            printf("Number of engines in stage %d? ", i+1);
        else
            printf("Number of engines? ");
        scanf("%d", &stages[i].engnum);
        if (stagenum > 1)
            printf("Maximum diameter of stage %d in inches? ", i + 1);
        else
            printf("Maximum body diameter in inches? ");
        scanf("%lE", &stages[i].maxd);

        if (i + 1 < stagenum) {
            printf("Staging delay for stage %d in seconds? ", i + 2);
            scanf("%lE", &stage_delay);
        }

               /* figure start and stop times for motor burn  and stage */
        if (i == 0)
            stages[i].start_burn = 0;
         else
            stages[i].start_burn = stages[i-1].end_stage;

            stages[i].end_burn = stages[i].start_burn + e_info[stages[i].engcnum].t2;
            stages[i].end_stage = stages[i].end_burn + stage_delay;

                  /* figure weight for stage and total rocket */


        stages[i].totalw = stages[i].weight +
                           (e_info[stages[i].engcnum].wt * stages[i].engnum);
        rocketwt += stages[i].totalw;

    }
    printf("Drag coefficient? ");
    scanf("%lE", &drag_coff);
    printf("Air temp in degF? ");
    scanf("%lE", &base_temp);
    base_temp = (base_temp - 32) * 5/9 + 273.15;  /* convert to degK */
    destnum = 0;
    do {
        printf("Send data to:\n\t(1) Screen\n\t(2) Printer\n\t");
        printf("(3) Disk file\nEnter #:");
        scanf("%d", &destnum);
#ifdef MSDOS
        if (destnum == 1)
            strcpy(fname, "CON");
        else if (destnum == 2)
            strcpy(fname, "PRN");
#endif
#ifdef UNIX
        if (destnum == 1)
            strcpy(fname, "/dev/tty");
        else if (destnum == 2)
            strcpy(fname, "/dev/lp");
#endif
#ifdef VMS
        if (destnum == 1)
            strcpy(fname, "TT:");
        else if (destnum == 2)
            strcpy(fname, "LP:");
#endif

        else {
            printf("Enter file name:");
            scanf("%s", fname);
        }
        if ((stream = fopen(fname,"w")) == NULL)
            printf("%s cannot be opened.\n",fname);
    } while (stream == NULL);
    wt = rocketwt;
    for (i = 0; i < stagenum; i++) {
      fprintf(stream,
              "%c\n%cStage   Engine            Bare     Launch  Dia.\n", ch1, ch1);
      fprintf(stream,
              "%cNumber  (Quant)Type       Wt(oz)   Wt(oz)  (in)      Cd\n", ch1);
      fprintf(stream,
              "%c------  ----------------  ------   ------  --------  ----\n", ch1);
      fprintf(stream,
              "%c   %d    (%d) %-12s   %.2lf     %.2lf     %.2lf    %.2lf\n",
              ch1, i + 1, stages[i].engnum, e_info[stages[i].engcnum].code,
              stages[i].weight / OZ2KG, wt / OZ2KG, stages[i].maxd, drag_coff);
      wt -= stages[i].totalw;
      print_engine_info(stages[i].engcnum);
    }
    if (verbose) {
      fprintf(stream, "%c\n%c%4s %11s %10s %12s %8s %8s %10s\n", ch1, ch1, 
              "Time", "Altitude", "Velocity",    "Acceleration", "Weight", "Thrust", "Drag");
      fprintf(stream, "%c%4s %10s %10s %12s %9s %9s %9s\n", ch1, 
              "(Sec)", "(Feet)", "(Feet/Sec)", "(Feet/Sec^2)", "(Ounces)", "(Newtons)" , "(Factor)");
      fprintf(stream, "%c%4s %10s %10s %10s %10s %10s %10s\n", ch1, 
              "-----", "------", "---------", "--------", "--------", "--------", "--------");
    } else fprintf(stream, "\n");
}
calc() {

    double floor();
    int j;                     /* iteration variable */
    int thrust_index = 0;     /* index into engine table */
    int stage_engcnum,           /* engcnum for current stage */
        this_stage = 0;        /* current stage */
    double stage_time = 0.0,     /* elapsed time for current stage */
           burn_time = 0.0,      /* Elapsed time for motor burn */
           stage_wt,             /* current wt of this_stage */
           stage_diam;              /* max diam at current time */
    double t_rod = 0.0, v_rod, x_rod; /* launch rod info */
    double c;
    double cc, drag_bias;
    double thrust = 0.0;          /* Thrust */
    double vel = 0.0;            /* Velocity */
    double accel = 0.0;          /* Acceleration */
    double alt = 0.0;            /* Altitude */
    double print_vel, print_alt, print_accel, print_mass;  /* used for printing */

    double mass = 0.0;
    double t = 0.0;
    double max_accel = 0.0;
    double max_vel = 0.0;
    double old_thrust = 0.0;   /* last engine thrust  from table */
    double old_time = 0.0;    /* last engine thrust time  from table */
    double d, m1;                 /* temp vars */

    int print_index = 0;              /* used for print loop */
    int launched = 0;            /* indicates rocket has lifted off */

    double delta_t = 0.001;        /* Time interval - 1ms */
    double r;              /* Constant used for drag calc */
    double air_density();

    stage_wt = stages[this_stage].totalw;
    stage_engcnum = stages[this_stage].engcnum;
    mass = rocketwt;

    for (j = 0; j < stagenum; j++) {
        stage_diam = stages[j].maxd;
        if (j > 0) {
            if (stage_diam < stages[j-1].maxd)
                stage_diam = stages[j-1].maxd;
        }
    }
    d = stage_diam * IN2M;
    c = r * PI * drag_coff * d * d * 0.125;

    for(;;) {
        /* Calculate decreasing air density */ 
        r = air_density (alt,base_temp); 
        c = r * PI * drag_coff * d * d * 0.125;

        t += delta_t;
        stage_time += delta_t;

                  /* handle staging, if needed */

        if ((t > stages[this_stage].end_stage) && (this_stage < stagenum-1)) {
            thrust_index = 0;
            old_thrust = 0.0;
            old_time = 0.0;
            mass = rocketwt -= stage_wt;
            this_stage++;
            stage_wt = stages[this_stage].totalw;
            stage_engcnum = stages[this_stage].engcnum;
            stage_time = burn_time = 0;
            for (j = this_stage; j < stagenum; j++) {
                stage_diam = stages[j].maxd;
                if (j > this_stage) {
                    if (stage_diam < stages[j-1].maxd)
                        stage_diam = stages[j-1].maxd;
                }
            }
            d = stage_diam * IN2M;
            c = r * PI * drag_coff * d * d * 0.125;
            fprintf(stream,"%cStage %d Ignition.\n", ch1, this_stage+1);
        }

                  /* Handle the powered phase of the boost */

        if ((t >= stages[this_stage].start_burn) &&
            (t <= stages[this_stage].end_burn)) {

               burn_time +=  delta_t;  /* add to burn time */

                /* see if we need to use the next thrust point */
          /* All times are relative to burn time for these calculations */

            if (burn_time > e_info[stage_engcnum].thrust[thrust_index]) {
                old_time = e_info[stage_engcnum].thrust[thrust_index];
                thrust_index++;
                old_thrust = e_info[stage_engcnum].thrust[thrust_index];
                thrust_index++;
            }

/*

Logic to smooth transision between thrust points.  Transisions are
linear rather than discontinuous.

*/

            thrust = e_info[stage_engcnum].thrust[thrust_index + 1] - old_thrust;
            thrust *= (burn_time - old_time) / (e_info[stage_engcnum].thrust[thrust_index] - old_time);
            thrust += old_thrust;
            thrust *= stages[this_stage].engnum;
            m1 = (e_info[stage_engcnum].m2 / e_info[stage_engcnum].t2) * delta_t;
            stage_wt -= m1;
            mass = rocketwt -= m1;
          } else thrust = 0;

/*

Crude approximations for MACH 1 Transition.
The drag bias will be applied to the subsonic value.
It will be equal to 1.0 for velocities less than MACH 0.8.
From MACH 0.8 to MACH 1.0 it will increase linearly to 2.0 times
the subsonic value.  From MACH 1.0 to MACH 1.2 it will decrease to
1.5 times the subsonic value.  Above MACH 1.2 it will remain at
a value of 1.5

*/
        if (vel < MACH0_8)
          drag_bias = 1;
        else if (vel < MACH1)
          drag_bias = (1.0 + ((vel - MACH0_8) / (MACH1 - MACH0_8)));
        else if (vel < MACH1_2)
          drag_bias = (2.0 - 0.5*((vel - MACH1) / (MACH1_2 - MACH1)));
        else
          drag_bias = 1.5;

        cc = c * drag_bias;

            /* Simple Newton calculations */

        accel = ((thrust - (cc * vel * vel)) / mass) - G;
        vel = vel + accel * delta_t;
        alt = alt + vel * delta_t;

            /* test for lift-off and apogee */

       if (vel > 0)
           launched = 1;                 /* LIFT-OFF */
       else if (!launched && (vel < 0))
           alt = vel = accel = 0;        /* can't fall off pad! */
       else if (launched && (vel < 0))
           break;                        /* apogee, all done */

        if(alt <= ROD) {
          t_rod = t;
          v_rod = vel * M2FT;
          x_rod = alt * M2FT;
        }

            /* do max evaluations */

        if (accel > max_accel) max_accel = accel;
        if (vel > max_vel) max_vel = vel;

            /* See if we need to print out the current values */

          if (print_index++  == 200) {   /* print every 200 ms */
              print_index = 0;
              print_alt = alt * M2FT;
              print_vel = vel * M2FT;
              print_accel = accel * M2FT;
              print_mass = mass / OZ2KG;
              if (verbose) fprintf(stream,"%4.1lf %10.1lf %10.1lf %10.1lf %10.3lf %10.3lf %10.3lf\n",
                  t + 0.01, print_alt, print_vel, print_accel, print_mass, thrust, drag_bias);
          }
   }
    print_alt = alt * M2FT;
    print_vel = vel * M2FT;
    print_accel = accel * M2FT;
    print_mass = mass / OZ2KG;
    fprintf(stream, "%cMaximum altitude attained was %.1lf feet (%.1lf meters).\n",
       ch1, print_alt, alt);
   fprintf(stream, "%cTime to peak altitude=%.2lf seconds.\n", ch1, t);
   fprintf(stream, "%cMaximum velocity was %.1lf feet/sec.\n", 
       ch1, max_vel * M2FT);
   fprintf(stream, "%cMaximum acceleration was %.1lf feet/sec^2.\n",
      ch1, max_accel * M2FT);
   fprintf(stream, "%cLaunch rod time = %.2lf,  altitude = %.1lf,  velocity = %.1lf\n",
           ch1, t_rod, x_rod, v_rod);
   fclose(stream);
}

main(argc, argv)
int argc;
char*argv[];
{
    char ans[3];

    if (argc > 1) verbose = FALSE;

    printf("\nRASP - Rocket Altitude Simulation Program V%s\n\n", VERSION);

    for (;;)
       {
        rocketwt = 0;
        nexteng = 0;
        choices();
        calc();
        printf("\nDo another one (Y/N)? ");
        scanf("%s", ans);
        if ((strncmp(ans,"y",1) == 0) ||
           (strncmp(ans,"Y",1) == 0))
         continue;
         else exit(0);
     }
}
